# Derived from https://gist.github.com/gormih/09d18e7da67271b79b6cb3537ebfa4f3

import datetime
import logging
import sqlite3

class SQLiteLoggingHandler(logging.Handler):
	def __init__(self, db = 'Log.db'):
		logging.Handler.__init__(self)
		self.db = db
		sql = """
			CREATE TABLE IF NOT EXISTS log (
				Created TEXT, Source TEXT, LogLevel INT, LogLevelName TEXT, Message TEXT,
				Args TEXT, Module TEXT, FuncName TEXT, LineNo INT, Exception TEXT,
				Process INT, Thread TEXT, ThreadName TEXT, TimeStamp INTEGER
			);"""
		with sqlite3.connect(self.db) as conn:
			conn.execute(sql)
			conn.commit()

	def format_time(self, record):
		record.dbtime = datetime.datetime.fromtimestamp(record.created).strftime("%Y-%m-%d %H:%M:%S.%f")
		record.timestamp = record.created

	def emit(self, record):
		self.format(record)
		self.format_time(record)
		if record.exc_info:  # for exceptions
			record.exc_text = logging._defaultFormatter.formatException(record.exc_info)
		else:
			record.exc_text = ""
		sql = """
			INSERT INTO log (
				Created, Source, LogLevel, LogLevelName, Message,
				Args, Module, FuncName, LineNo, Exception, 
				Process, Thread, ThreadName, TimeStamp
			)
			VALUES (
				?, ?, ?, ?, ?, 
				?, ?, ?, ?, ?, 
				?, ?, ?, ?
			);"""
		try:
			with sqlite3.connect(self.db) as conn:
				conn.execute(sql, (
					record.dbtime, record.name, record.levelno, record.levelname, record.msg, 
					str(record.args), record.module, record.funcName, record.lineno, record.exc_text, 
					record.process, record.thread, record.threadName, record.timestamp))
				conn.commit()  # not efficient, but hopefully thread-safe
		except Exception as e:
			print("Error in logging:", e)
	
	def get_last_message(self):
		with sqlite3.connect(self.db) as conn:
			return conn.execute("SELECT Message FROM log ORDER BY TimeStamp DESC LIMIT 1;").fetchone()


import logging

class ColoredLoggingFormatter(logging.Formatter):

    def __init__(self, formatString = None):
        defaultFormat = "[%(asctime)s] %(levelname)s [%(name)s.%(funcName)s:%(lineno)d] %(message)s"
        self.formatString = formatString if formatString else defaultFormat
        self.grey = "\x1b[38;20m"
        self.yellow = "\x1b[33;20m"
        self.red = "\x1b[31;20m"
        self.bold_red = "\x1b[31;1m"
        self.reset = "\x1b[0m"

    def getFormats(self):
        return {
            logging.DEBUG: self.grey + self.formatString + self.reset,
            logging.INFO: self.grey + self.formatString + self.reset,
            logging.WARNING: self.yellow + self.formatString + self.reset,
            logging.ERROR: self.red + self.formatString + self.reset,
            logging.CRITICAL: self.bold_red + self.formatString + self.reset
        }

    def format(self, record):
        log_fmt = self.getFormats().get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def write_text(folder : str, file : str, text : str):
	write_or_append_text(folder, file, text, "w")

def append_text(folder, file, text):
	write_or_append_text(folder, file, text, "a")

def write_or_append_text(folder, file, text, mode):
	from pathlib import Path
	Path(folder).mkdir(parents = True, exist_ok = True)
	f = open(folder + "/" + file, mode)
	f.write(text)
	f.close()



def fetch_json_key_value(file_name, key, defaultValue = None):
	import json
	import os.path
	if not os.path.isfile(file_name):
		return defaultValue
	with open(file_name) as settingsFile:
		settings = json.load(settingsFile)
		return settings.get(key, defaultValue)



import time
import unittest
import uuid

from easai.utils.local_cache import LocalCache

class LocalCacheTests(unittest.TestCase):

	def test_getOrAdd(self):
		cache = LocalCache("output/LocalCacheTest.db")
		key = "test-" + str(uuid.uuid4())
		class mock:
			def __init__(self):
				self.wasCalled = False
			def getValue(self):
				self.wasCalled = True
				return "42"
		
		myMock = mock()
		value = cache.get_or_add(key, myMock.getValue)
		self.assertEqual(value, "42")
		self.assertEqual(myMock.wasCalled, True)
		myMock = mock()
		value = cache.get_or_add(key, myMock.getValue)
		self.assertEqual(value, "42")
		self.assertEqual(myMock.wasCalled, False)

	def test_getOrAddWithLifetime(self):
		cache = LocalCache("output/LocalCacheTest.db")
		key = "test-" + str(uuid.uuid4())
		value = cache.get_or_add_with_lifetime(key, lambda: "1", 0.5)
		self.assertEqual(value, "1")
		value = cache.get_or_add_with_lifetime(key, lambda: "2", 0.5)
		self.assertEqual(value, "1")
		time.sleep(1)
		value = cache.get_or_add_with_lifetime(key, lambda: "2", 0.5)
		self.assertEqual(value, "2")



import datetime
import logging
import unittest

from easai.utils.sqlite_logging_handler import SQLiteLoggingHandler

class SQLiteLoggingHandlerTests(unittest.TestCase):
	def test_logger(self):
		sqlite_handler = SQLiteLoggingHandler(db = "output/log.db")
		sqlite_handler.setLevel(logging.DEBUG)
		logger = logging.getLogger()
		logger.addHandler(sqlite_handler)
		log_message = "testing: " + str(datetime.datetime.now())

		logger.error(log_message)

		last = sqlite_handler.get_last_message()
		self.assertEqual(last[0], log_message)



from typing import Callable

class AssistantToolParameter:
	def __init__(self, name: str, description: str, type: str = "string", enum: list[str] = None):
		self.name = name
		self.description = description
		self.type = type
		self.enum = enum
	
	def get_open_ai_tool_properties(self) -> dict:
		value = {
			"type": self.type,
			"description": self.description
		}
		if self.enum:
			value["enum"] = self.enum
		return value

class AssistantTool:
	def __init__(self, method: Callable, description: str, parameters: list[AssistantToolParameter] = [], name: str = None):
		self.parameters: list[AssistantToolParameter] = parameters
		self.description = description
		self.method: Callable = method
		self.name = name if name else method.__name__
	
	def map_to_open_ai_tool(self) -> dict:
		properties = {}
		for parameter in self.parameters:
			properties[parameter.name] = parameter.get_open_ai_tool_properties()
		return {
			"type": "function",
			"function": {
				"name": self.name,
				"description": self.description,
				"parameters": {
					"type": "object",
					"properties": properties
				}
			}
		}

def create_tool_from_method(method: Callable) -> AssistantTool:
	method_documentation = method.__doc__
	description = method_documentation.split("\n")[0] if method_documentation else method.__name__
	parameter_docs = method_documentation.split("\n")[1:] if method_documentation else []
	method_parameters = method.__code__.co_varnames

	def get_parameter_description(parameter_name: str) -> tuple:
		for parameter_doc in parameter_docs:
			if parameter_name in parameter_doc:
				parameter_description = parameter_doc.split(":")[1].strip()
				parameter_type = "integer" if "(int)" in parameter_doc else "string"
				return parameter_name, parameter_description, parameter_type
		return parameter_name, parameter_name, "string"

	parameter_descriptions = [get_parameter_description(name) for name in method_parameters if name != "self"]
	parameters = [AssistantToolParameter(
		name = parameter_description[0], description = parameter_description[1], type = parameter_description[2]) for parameter_description in parameter_descriptions]
	return AssistantTool(
		method = method,
		description = description,
		parameters = parameters
	)



from easai.assistant.log import AiLogBase, LoggingAiLog
from easai.assistant.assistant import Assistant, run_assistant, get_system_prompt, get_user_prompt
from easai.utils.console_formatting import background, foreground, style

def run_chat_console(
		assistant: Assistant = None,
		your_name: str = "You",
		ai_logger: AiLogBase = LoggingAiLog()):
	assistant = assistant if assistant is not None else Assistant()
	messages = [get_system_prompt(assistant.system_prompt)]
	print_assistant_message("Hello! How can I help you?")
	while True:
		prompt = input(background.BLUE + foreground.WHITE + get_role_console_line(your_name) + " ")
		print()
		if prompt == "bye":
			print_assistant_message("Good bye!", skip_bye_info = True)
			return
		messages.append(get_user_prompt(prompt))
		messages = run_assistant(
			assistant = assistant,
			messages = messages,
			assistant_message_callback = print_assistant_message,
			tool_message_callback = print_tool_message,
			ai_logger = ai_logger)

def get_role_console_line(role : str):
	return "  " + role.ljust(11) + style.RESET_ALL + " : "

def print_assistant_message(content, skip_bye_info = False):
	bye_info = " (Enter 'bye' to quit.)" if not skip_bye_info else ""
	print(background.GREEN + foreground.WHITE + get_role_console_line("Assistant"), content, bye_info)
	print()

def print_tool_message(content):
	print(background.YELLOW + foreground.WHITE + get_role_console_line("Tool"), content)
	print()



import unittest

from easai.assistant.tool import create_tool_from_method

class AssistantToolUnitTests(unittest.TestCase):

	def test_create_tool_from_method(self):
		def my_method(param1, param2):
			"""This is my method documentation
			Args:
				param1 (int): The first parameter
				param2 (str): The second parameter
			"""
			pass

		tool = create_tool_from_method(my_method)

		self.assertEqual(tool.name, "my_method")
		self.assertEqual(tool.description, "This is my method documentation")
		self.assertEqual(tool.parameters[0].name, "param1")
		self.assertEqual(tool.parameters[0].description, "The first parameter")
		self.assertEqual(tool.parameters[0].type, "integer")
		self.assertEqual(tool.parameters[1].name, "param2")
		self.assertEqual(tool.parameters[1].description, "The second parameter")
		self.assertEqual(tool.parameters[1].type, "string")

	def test_create_tool_from_method_without_docs(self):
		def my_method(param1, param2):
			pass

		tool = create_tool_from_method(my_method)

		self.assertEqual(tool.name, "my_method")
		self.assertEqual(tool.description, "my_method")
		self.assertEqual(tool.parameters[0].name, "param1")
		self.assertEqual(tool.parameters[0].description, "param1")
		self.assertEqual(tool.parameters[0].type, "string")
		self.assertEqual(tool.parameters[1].name, "param2")
		self.assertEqual(tool.parameters[1].description, "param2")
		self.assertEqual(tool.parameters[1].type, "string")



import unittest

from easai.assistant.assistant import Assistant, get_user_prompt, run_assistant
from easai.assistant.tool import AssistantTool, AssistantToolParameter

class AssistantSystemTests(unittest.TestCase):
	def test_run_assistant(self):
		from dotenv import load_dotenv
		load_dotenv(override = True)
		messages = [get_user_prompt("How is the weather in Italy?")]
		self.tool_called = False
		def get_weather(country: str):
			return "sunny in " + country
		def assert_tool_call(message):
			self.tool_called = True
			self.assertEqual(message, "Calling function get_weather with {'country': 'Italy'}")
		tools = [
			AssistantTool(
				method = get_weather,
				description = "Retrieve the current weather",
				parameters = [
					AssistantToolParameter(name = "country", description = "The country")
				])
		]
		
		result = run_assistant(
			assistant = Assistant(tools = tools),
			messages = messages,
			tool_message_callback = assert_tool_call)
		
		self.assertIn("sunny", result[-1].content)
		self.assertTrue(self.tool_called)

	def test_run_assistant_without_tools(self):
		from dotenv import load_dotenv
		load_dotenv(override = True)
		messages = [get_user_prompt("Explain the sun")]
		
		result = run_assistant(messages = messages)
		
		self.assertIn("sun", str.lower(result[-1].content))



import json
import logging
import sqlite3
from datetime import datetime

from openai.types.chat.chat_completion_message import ChatCompletionMessage

class AiLogBase:
	def log(self, model, promptTokens, completionTokens, messages):
		messagesString = json.dumps([self.extract_message_values(message) for message in messages])
		self.do_log(model, promptTokens, completionTokens, messagesString)

	def do_log(self, model, promptTokens, completionTokens, messagesString):
		pass

	def extract_message_values(self, message):
		if isinstance(message, ChatCompletionMessage):
			return {
				"role": message.role, 
				"content": message.content,
				"tool_calls": [self.extract_tool_call_values(toolCall) for toolCall in message.tool_calls] if message.tool_calls else None
			}
		return {
			"role": message["role"],
			"content": message["content"],
			"name": message["name"] if "name" in message else None,
			"tool_call_id": message["tool_call_id"] if "tool_call_id" in message else None
		}
	
	def extract_tool_call_values(self, toolCall):
		return {
			"id": toolCall.id,
			"functionName": toolCall.function.name, 
		  	"arguments": toolCall.function.arguments
		}

class LoggingAiLog(AiLogBase):
	def __init__(self):
		self.logger = logging.getLogger(__name__)

	def do_log(self, model, promptTokens, completionTokens, messages_string):
		self.logger.debug("Model: %s. Prompt Tokens: %d. Completion Tokens: %d. Messages: %s", model, promptTokens, completionTokens, messages_string)

class SqliteAiLog(AiLogBase):
	def __init__(self, database: str):
		self.db = sqlite3.connect(database)
		self.db.execute("CREATE TABLE IF NOT EXISTS AiLog (AiLogId INTEGER PRIMARY KEY, Date TEXT, Model TEXT, PromptTokens INTEGER, CompletionTokens INTEGER, Messages TEXT, TimeStamp INTEGER)")

	def do_log(self, model, promptTokens, completionTokens, messages_string):
		now = datetime.now()
		timeStamp = int(now.timestamp())
		isoString = now.isoformat()
		self.db.execute(
			"INSERT INTO AiLog (AiLogId, Date, Model, PromptTokens, CompletionTokens, Messages, TimeStamp) VALUES (?, ?, ?, ?, ?, ?, ?)", 
			(None, isoString, model, promptTokens, completionTokens, messages_string, timeStamp))
		self.db.commit()



import json
import logging

from openai import OpenAI

from easai.assistant.tool import AssistantTool
from easai.assistant.log import AiLogBase, LoggingAiLog

class Assistant:
	def __init__(self, client: OpenAI = None, model: str = "gpt-4o-mini", 
			  system_prompt: str = "You are a helpful assistant.",
			  tools: list[AssistantTool] = [], max_iterations: int = 10):
		self.client: OpenAI = client if client is not None else OpenAI()
		self.model: str = model
		self.system_prompt: str = system_prompt
		self.tools: list[AssistantTool] = tools
		self.max_iterations: int = max_iterations

def run_assistant(
		messages: list = [],
		assistant: Assistant = None,
		assistant_message_callback = None,
		tool_message_callback = None,
		ai_logger: AiLogBase = LoggingAiLog()) -> list:
	"""Run assistant, calling tools until satisfied.

	Args:
		assistant (Assistant): A helpful assistant
		messages (list): List of messages
		assistant_message_callback (function, optional): Function to call when the assistant message is received. Defaults to None.
	"""
	assistant = assistant if assistant is not None else Assistant()
	client_tools = [tool.map_to_open_ai_tool() for tool in assistant.tools]
	tool_methods = {tool.name: tool.method for tool in assistant.tools}
	messages.insert(0, get_system_prompt(assistant.system_prompt))
	message_count_at_last_log = len(messages) - 1
	for _ in range(assistant.max_iterations):
		chat_completion = assistant.client.chat.completions.create(
			messages = messages,
			model = assistant.model,
			tools = client_tools if len(client_tools) > 0 else None
		)
		choice = chat_completion.choices[0]
		finish_reason = choice.finish_reason
		message = choice.message
		messages.append(message)
		ai_logger.log(assistant.model, chat_completion.usage.prompt_tokens, chat_completion.usage.completion_tokens, messages[message_count_at_last_log:])
		message_count_at_last_log = len(messages)
		if finish_reason == "stop":
			if assistant_message_callback:
				assistant_message_callback(message.content)
			return messages
		elif finish_reason == "tool_calls":
			for tool_call in message.tool_calls:
				call_function = tool_call.function
				if call_function.name in tool_methods:
					function_name = call_function.name
					function_args = json.loads(call_function.arguments)
					function_response = None
					tool_method = tool_methods[function_name]
					if function_args:
						if tool_message_callback:
							tool_message_callback("Calling function " + function_name + " with " + str(function_args))
						function_response = tool_method(**function_args)
					else:
						if tool_message_callback:
							tool_message_callback("Calling function " + function_name)
						function_response = tool_method()
					messages.append({
						"role": "tool",
						"name": function_name,
						"tool_call_id": tool_call.id,
						"content": get_limited_message_content(json.dumps(function_response))
					})
	logger = logging.getLogger(__name__)
	logger.error("ERROR: Assistant exited without stop reason, most likely due to too many iterations.")
	return messages

def get_limited_message_content(content, max_message_content_length = 100 * 1000):
	if len(content) <= max_message_content_length:
		return content

	logger = logging.getLogger(__name__)
	logger.warning(
		"Message content %s was too long, truncating to %d characters.",
		content[:50] + " ... " + content[-50:],
		max_message_content_length)
	return content[:max_message_content_length] + "..."

def get_system_prompt(content, max_message_content_length = 100 * 1000):
	return get_prompt("system", content, max_message_content_length)

def get_user_prompt(content, max_message_content_length = 100 * 1000):
	return get_prompt("user", content, max_message_content_length)

def get_prompt(role, content, max_message_content_length = 100 * 1000):
	return { "role": role, "content": get_limited_message_content(content, max_message_content_length) }



import unittest

from easai.assistant.chat_console import run_chat_console

class ChatConsoleTests(unittest.TestCase):
	def test_run_chat_console(self):
		from dotenv import load_dotenv
		load_dotenv(override = True)
		run_chat_console()



class LocalCache:
	def __init__(self, cache_db_path = "output/local_cache.db"):
		import sqlite3
		self.db = sqlite3.connect(cache_db_path)

	def get_or_add(self, key, getValue):
		self.db.execute("CREATE TABLE IF NOT EXISTS Cache (key TEXT PRIMARY KEY, value TEXT)")
		existing_value = self.db.execute("SELECT value FROM Cache WHERE key = ?", (key,)).fetchone()
		if existing_value:
			return existing_value[0]
		
		value = getValue()
		self.db.execute("INSERT OR REPLACE INTO Cache (key, value) VALUES (?, ?)", (key, value))
		self.db.commit()
		return value
	
	def get_or_add_with_lifetime(self, key, getValue, lifetimeSeconds):
		import json, time
		now_timestamp = time.time()
		must_be_created_after_timestamp = now_timestamp - lifetimeSeconds
		self.db.execute("CREATE TABLE IF NOT EXISTS CacheWithLifetime (Key TEXT PRIMARY KEY, Value TEXT, Timestamp INTEGER)")
		existing_value = self.db.execute("SELECT Value FROM CacheWithLifetime WHERE Key = ? AND Timestamp > ?", (key,must_be_created_after_timestamp,)).fetchone()
		if existing_value:
			return json.loads(existing_value[0])
		
		value = getValue()
		self.db.execute(
			"INSERT OR REPLACE INTO CacheWithLifetime (Key, Value, Timestamp) VALUES (?, ?, ?)",
			(key, json.dumps(value), now_timestamp))
		self.db.commit()
		return value
